import datetime as datetime
from dataclasses import dataclass, field
from pathlib import Path
import numpy as np
from dateutil import parser
import pandas as pd
from io import StringIO
from scidatapy.scidataclasses.gen import SciDataFile, ProcessedSciData
import scidatapy.utils as sdu
import multiprocessing as mp
import pint
import xmltodict


@dataclass(order=False)
class MeasurementData:
    measurement_details: list = field(init=False, compare=False)
    machine_code: str = field(init=False, compare=False)
    software_details: str = field(init=False, compare=False)
    software_version: any = field(init=False, compare=False)
    measurement_datetime: datetime.datetime = field(init=False, compare=False)


@dataclass(order=False)
class MTSData(SciDataFile, MeasurementData):

    def load_data(self, file_path):
        # Specific to File
        self.data_file_path = file_path
        self.n_head_lines = 5
        self.n_tail_lines = 0
        self.dlim_white_sp = True
        self.cols_from_topline = None
        # parse with generic parser
        self.parser()
        # gather info from header specific to File
        self.machine_code = self.header[0].strip()
        self.measurement_details = self.header[2].strip().split('\t\t\t')
        self.raw_data.cols = self.header[3].strip().split('\t')
        self.raw_data.col_units = self.header[4].strip().split('\t')
        test_datetime_str = self.measurement_details[-1].split('\t')[-1]
        self.measurement_datetime = parser.parse(test_datetime_str)
        self.raw_data.df.rename(columns=dict(zip(self.raw_data.df.columns, self.raw_data.cols)), inplace=True)
        # Identify time column, get the raw sample frequency
        self.raw_data.time_col_name = 'Time'
        self.raw_data.get_freq()

    def plot_mts_raw(self):
        self.raw_data.df.plot(x=0, subplots=True)


def load_mts_data(file_path):
    mts_data = MTSData()
    mts_data.load_data(file_path)
    return mts_data


@dataclass(order=False)
class AEWinData(SciDataFile, MeasurementData):
    o_file_path: Path() = field(init=False, compare=False)

    def load_data(self, file_path):
        # Specific to File
        self.data_file_path = file_path
        self.n_head_lines = 6
        self.n_tail_lines = 0
        self.dlim_white_sp = True
        # self.cols_from_topline = 'infer'
        # parse with generic parser
        self.parser()
        # gather info from header specific to file
        self.o_file_path = Path(self.header[0].strip())
        self.software_details = self.header[1].strip()
        self.software_version = self.header[3].strip()
        # self.raw_data.cols = self.raw_data.df.columns.tolist()
        # TODO: get units from the mistras aewin software
        # self.raw_data.col_units =
        # self.measurement_details =
        test_datetime_str = self.header[-1].strip()
        self.measurement_datetime = parser.parse(test_datetime_str)
        # Identify time column, get the raw sample frequency
        self.raw_data.time_col_name = 'SSSSSSSS.mmmuuun'
        self.raw_data.get_freq()


@dataclass(order=False)
class RTecTriboSteps(ProcessedSciData):
    acc_col_name: any = field(init=False, compare=False)
    vel_col_name: any = field(init=False, compare=False)


@dataclass(order=False)
class RTecTriboData(SciDataFile, MeasurementData):
    setup_df: pd.DataFrame() = field(init=False, compare=False)
    test_timedelta: datetime.timedelta = field(init=False, compare=False)
    processed_data: ProcessedSciData = field(init=False, compare=False, default_factory=ProcessedSciData)
    n_steps: any = field(init=False, compare=False)
    time_units: any = field(init=False, compare=False)
    steps_data: RTecTriboSteps = field(init=False, compare=False, default_factory=RTecTriboSteps)
    # remove below..
    # steps_df: pd.DataFrame() = field(init=False, compare=False)
    # acc_col_name: any = field(init=False, compare=False)
    # vel_col_name: any = field(init=False, compare=False)

    def load_data(self, file_path):
        # Specific to File
        self.data_file_path = file_path
        self.n_head_lines = 2
        self.n_tail_lines = 2
        # self.dlim_white_sp = False
        # self.cols_from_topline = 'infer'
        # parse with generic parser
        self.parser()
        self.setup_df = pd.read_csv(StringIO('\n'.join(self.header)), index_col=False)
        timedelta_str = self.tail[1].split(',')[0].strip()
        timedelta_dt = parser.parse(timedelta_str)
        self.test_timedelta = timedelta_dt.combine(timedelta_dt.min, timedelta_dt.time()) - timedelta_dt.min
        self.raw_data.col_units = sdu.units_rsplit_cols(self.raw_data.df.columns)
        self.raw_data.time_col_name = ' Timestamp'.strip()
        self.raw_data.get_freq()

    def process_raw(self):
        self.processed_data.df = self.raw_data.df
        # take out some columns - unnamed and na
        unnamed_cols = self.processed_data.df.columns.str.contains('unnamed', case=False)
        na_cols = self.processed_data.df.isna().all().values
        self.processed_data.df.drop(columns=self.processed_data.df.columns[unnamed_cols & na_cols].to_list(),
                                    inplace=True)
        # self.processed_data.df = self.processed_data.df.loc[:, ~(unnamed_cols & na_cols)]
        self.processed_data.col_units = sdu.units_rsplit_cols(self.processed_data.df.columns)
        # self.processed_data.col_units = [i for i, j in zip(self.raw_data.col_units, ~(unnamed_cols & na_cols)) if j]
        # Time column units have not been added?
        self.processed_data.time_col_name = self.raw_data.time_col_name
        time_col = self.processed_data.df.columns.get_loc(self.processed_data.time_col_name)
        self.time_units = 's'
        self.processed_data.col_units[time_col] = self.time_units
        # Other column units fixes..
        self.processed_data.col_units = sdu.units_col_fixes(self.processed_data.col_units)
        # Quantize it
        self.processed_data.df_ = sdu.pint_df(self.processed_data.df, self.processed_data.col_units)
        # Get steps info
        # self.n_steps = self.processed_data.df['RecipeStep'].iloc[-1]
        self.n_steps = len(self.processed_data.df['RecipeStep'].unique())
        self.steps_data.df = pd.DataFrame()
        # self.steps_df['RecipeStep'] = np.arange(self.n_steps) + 1
        # self.steps_data.df['RecipeStep'] = self.processed_data.df['RecipeStep'].astype(int).unique()
        self.steps_data.df['RecipeStep'] = self.processed_data.df['RecipeStep'].unique()
        df_step_grps = self.processed_data.df.groupby('RecipeStep')
        self.steps_data.df['pStartLoc'] = df_step_grps[self.processed_data.time_col_name].idxmin().astype(int).values
        self.steps_data.df['pStopLoc'] = df_step_grps[self.processed_data.time_col_name].idxmax().astype(int).values
        self.steps_data.col_units = ['', '', '']
        self.steps_data.df_ = sdu.pint_df(self.steps_data.df, self.steps_data.col_units)

    def get_recipe_info(self, recipe_file_path):
        with open(recipe_file_path) as fd:
            doc = xmltodict.parse(fd.read())

        def find_in_step_detail(detail, step_num, xml_doc=doc):
            detail_value = 0
            detail_desc = ''
            step_details = xml_doc['TribologyRecipeSummary']['Steps'][step_num]['StepDetail'].rsplit('\n')
            step_detail = sdu.find_str_in_list(step_details, detail)
            if len(step_detail) == 1:
                detail_and_val = step_detail[0].rsplit(':')
                detail_desc = detail_and_val[0].strip()
                # TODO: when the detail values are changing during step (currently code works for constant value)
                # check for something like this Fz:30 --&gt; 60 N
                detail_value = detail_and_val[1].strip()
                if len(detail_value) > 0:
                    detail_value = float(detail_value)
                    # find the direction if it is velocity
                    if detail == 'Velocity':
                        direction_detail = sdu.find_str_in_list(step_details, 'Direction')
                        direction_and_val = direction_detail[0].rsplit(':')
                        # direction_desc = direction_and_val[0].strip()
                        direction_value = direction_and_val[1].strip()
                        if direction_value == 'CCW':
                            detail_value = -detail_value
            return detail_desc, detail_value

        def add_detail_column(detail, n_steps=self.n_steps, steps_df=self.steps_data.df):
            n_steps = int(n_steps)
            detail_col_name = ''
            detail_values = np.zeros(n_steps)
            # for step in range(0, n_steps):
            for istep in steps_df.index.tolist():
                step = steps_df.loc[istep, 'RecipeStep']
                detail_desc, detail_value = find_in_step_detail(detail, int(step - 1))
                if abs(detail_value) > 0:
                    if len(detail_col_name) == 0:
                        detail_col_name = detail_desc
                    detail_values[istep] = detail_value
            if len(detail_col_name) > 0:
                steps_df[detail_col_name] = detail_values
            return steps_df, detail_col_name

        self.steps_data.df, self.steps_data.vel_col_name = add_detail_column('Velocity')
        self.steps_data.df, self.steps_data.acc_col_name = add_detail_column('Acceleration')
        self.steps_data.col_units = sdu.units_rsplit_cols(self.steps_data.df.columns)
        self.steps_data.col_units = sdu.units_col_fixes(self.steps_data.col_units)
        self.steps_data.df_ = sdu.pint_df(self.steps_data.df, self.steps_data.col_units)

    def add_time_col(self):
        n_st_tcs = 0
        # + ' (' + self.time_units + ')'
        self.steps_data.df['StopTime'] = self.processed_data.df.groupby('RecipeStep')[
            self.processed_data.time_col_name].max().values
        n_st_tcs += 1

        # extra_for_acc = self.processed_data.df.groupby('RecipeStep')['Timestamp'].apply(lambda x: any(x.diff() <= 0))
        # Make a cumulative time column
        # if (len(self.acc_col_name) > 0) & (any(extra_for_acc)):
        #     self.steps_df['TotalTime'] = (self.steps_df['StopTime'] +
        #                                   (extra_for_acc.values.astype(int)*self.steps_df[self.acc_col_name])).cumsum()
        if len(self.steps_data.acc_col_name) > 0:
            self.steps_data.df['TotalTime'] = (
                    self.steps_data.df['StopTime'] + self.steps_data.df[self.steps_data.acc_col_name]).cumsum()
        else:
            self.steps_data.df['TotalTime'] = self.steps_data.df['StopTime'].cumsum()
        n_st_tcs += 1

        split_locs = self.steps_data.df['pStartLoc'].iloc[1:].values
        df_split = np.array_split(self.processed_data.df, split_locs)

        self.steps_data.df['rStartTime'] = 0.0

        def add_cumtime_col(df, steps_df=self.steps_data.df, acc_col=self.steps_data.acc_col_name,
                            time_col=self.processed_data.time_col_name):
            same_time_loc = steps_df.index[steps_df['RecipeStep'] == df['RecipeStep'].iloc[0]][0]
            if len(acc_col) > 0:
                acc_time = steps_df[acc_col].loc[same_time_loc]
            else:
                acc_time = 0.0
            acc_end_loc = -1
            if (acc_time > 0) & (any(df[time_col].diff() <= 0)):
                acc_end_loc = (df[time_col] - acc_time).abs().idxmin()
            if df['RecipeStep'].iloc[0] > steps_df.loc[0, 'RecipeStep']:
                # time_offset = steps_df['TotalTime'].iloc[same_time_loc - 1]
                time_offset = steps_df.loc[same_time_loc - 1, 'TotalTime']
            else:
                time_offset = 0.0
                # self.steps_data.df['rStartTime'] = 0.0
            df['Time'] = df[time_col] + time_offset
            if acc_end_loc >= 0:
                # df['Time'].loc[(acc_end_loc + 1):] += acc_time
                # self.steps_data.df['rStartTime'].iloc[same_time_loc] = df['Time'].loc[acc_end_loc + 1]
                df.loc[(acc_end_loc + 1):, 'Time'] += acc_time
                self.steps_data.df.loc[same_time_loc, 'rStartTime'] = df.loc[acc_end_loc + 1, 'Time']
            else:
                # self.steps_data.df['rStartTime'].iloc[same_time_loc] = df['Time'].iloc[0]
                self.steps_data.df.loc[same_time_loc, 'rStartTime'] = df['Time'].iloc[0]
            return df
            # return df[self.processed_data.time_col_name] + time_offset

        result_chunks = [add_cumtime_col(chunk) for chunk in df_split]
        n_st_tcs += 1
        self.steps_data.col_units += [self.time_units] * n_st_tcs
        self.steps_data.df_ = sdu.pint_df(self.steps_data.df, self.steps_data.col_units)

        result_df = pd.concat(result_chunks)
        self.processed_data.df = result_df
        # time_units = self.processed_data.df_[self.processed_data.time_col_name].iloc[0].units
        time_units = self.time_units
        self.processed_data.df_['Time'] = pint.PintArray(self.processed_data.df['Time'].values, dtype=time_units)
        self.processed_data.time_col_name = 'Time'
        self.processed_data.df, self.processed_data.col_units = sdu.un_pint_df(self.processed_data.df_)
