import pandas as pd
from scipy import fftpack
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
import pint
import numpy as np
from matplotlib import pyplot as plt
# from multiprocessing import Pool
from matplotlib.backends.backend_pdf import PdfPages


def pint_df(df, col_units):
    df_ = df.copy()
    df_.columns = pd.MultiIndex.from_tuples(list(zip(df_.columns, col_units)))
    df_ = df_.pint.quantify(level=-1)
    return df_


def un_pint_df(df_):
    df = df_.copy()
    df = df.pint.dequantify()
    col_units = df.columns.get_level_values(1).tolist()
    df.columns = df.columns.get_level_values(0)
    return df, col_units


def pint_abs_col(df_, col_name, add_col=True, new_col_name=''):
    df_d = df_.copy()
    arr = df_d[col_name].pint.quantity.magnitude
    pint_abs = pint.PintArray(np.abs(arr), dtype=df_d[col_name].pint.units)
    if add_col:
        if len(new_col_name) == 0:
            new_col_name = 'abs.' + col_name
        df_d[new_col_name] = pint_abs
        return df_d
    else:
        return pint_abs


def pint_df_col_der(df_, col_name, time_col_name, nder=1, window=7, polyorder=2, add_col=True, new_col_name=''):
    df_d = df_.copy()
    # arr = df_d[col_name].pint.magnitude.to_numpy()
    # arr = df_d[col_name].pint.quantity.magnitude
    # arr_ycol = arr
    df, df_units = un_pint_df(df_)
    der = df_col_sm_or_der(df, col_name, time_col_name, nder, window, polyorder, add_col=False)
    # arr, xt, uni_time_col = var2uni_or_xt_col(df, col_name, time_col_name)
    # der = savgol_filter(arr, window_length=window, polyorder=polyorder, deriv=nder)
    # if not uni_time_col:
    #     f_der = interp1d(xt, der)
    #     der = f_der(df[time_col_name].values)
    del df, df_units
    pint_der = pint.PintArray(der, dtype=df_d[col_name].pint.units / (df_d[time_col_name].pint.units ** nder))
    if add_col:
        if len(new_col_name) == 0:
            if nder > 1:
                new_col_name = 'd' + str(nder) + col_name
            elif nder == 0:
                new_col_name = col_name + '_sm'
            else:
                new_col_name = 'd' + col_name
        df_d[new_col_name] = pint_der
        return df_d
    else:
        return pint_der


def df_col_sm_or_der(df, col_name, time_col_name, nder=0, window=51, polyorder=2, add_col=True, new_col_name='',
                     interp=True):
    window = int(round_up_to_odd(window))
    arr, xt, uni_time_col = var2uni_or_xt_col(df, col_name, time_col_name)
    if uni_time_col | interp:
        der = savgol_filter(arr, window_length=window, polyorder=polyorder, deriv=nder)
        xtder = (xt[-1] - xt[-2])
        der = der / (xtder ** nder)
        if not uni_time_col:
            f_der = interp1d(xt, der)
            der = f_der(df[time_col_name].values)
    else:
        arr = df[col_name].values
        xt = df[time_col_name].values
        der = savgol_filter(arr, window_length=window, polyorder=polyorder, deriv=nder)
        xtder = 1
        if nder > 0:
            xtder = savgol_filter(xt, window_length=window, polyorder=polyorder, deriv=1)
        der = der / (xtder ** nder)
    if add_col:
        if len(new_col_name) == 0:
            if nder > 1:
                new_col_name = 'd' + str(nder) + col_name
            elif nder == 0:
                new_col_name = col_name + '_sm'
            else:
                new_col_name = 'd' + col_name
        df[new_col_name] = der
        return df
    else:
        return der


def round_up_to_odd(f):
    return np.ceil(f) // 2 * 2 + 1


def find_changes(der2, thresh=0.75, window=7):
    max_der2 = np.max(np.abs(der2))
    # large = np.where(np.abs(der2) > max_der2/2)[0]
    large = np.where(np.abs(der2) > thresh * max_der2)[0]
    gaps = np.diff(large) > window
    begins = np.insert(large[1:][gaps], 0, large[0])
    ends = np.append(large[:-1][gaps], large[-1])
    changes = ((begins + ends) / 2).astype(np.int)
    return changes


def var2uni_or_xt_col(df, col_name, time_col_name, uni_time_col=False, xt=None):
    if xt is None:
        xt = []
    if uni_time_col:
        return df[col_name].values, df[time_col_name].values, uni_time_col
    y_col = []
    diff = df[time_col_name].diff()
    sample_freq = diff.mode()[0]
    diff_min = diff.min()
    diff_max = diff.max()
    if (abs(diff_min - diff_max) < 1e-7) and (len(xt) == 0):
        xt = df[time_col_name].values
        y_col = df[col_name].values
        uni_time_col = True
    # elif (sample_freq * diff_min > 0) and (sample_freq * diff_max > 0):
    else:
        f_col = interp1d(df[time_col_name].values, df[col_name].values)
        if len(xt) == 0:
            # assuming increasing...
            # t_begin = df[time_col_name].iloc[0]
            # t_end = df[time_col_name].iloc[-1]
            t_begin = df[time_col_name].min()
            t_end = df[time_col_name].max()
            xt = np.linspace(t_begin, t_end, len(df))
        y_col = f_col(xt)
    return y_col, xt, uni_time_col


def units_rsplit_cols(cols, rs='()'):
    col_units = cols.str.rsplit(rs[0]).str.get(1).str.rsplit(rs[1]).str.get(0).str.strip().to_list()
    return col_units


def units_col_fixes(col_units):
    col_units = ['' if (str(x) == 'nan' or x == 'nan') else x for x in col_units]
    col_units = [('deg' + x[1:]) if (len(x) > 0 and x[0] == '°') else x for x in col_units]
    return col_units


def find_str_in_list(strs_list, str_to_find):
    str_list = []
    str_list = [s for s in strs_list if str_to_find in s]
    return str_list


def fft_this_signal(sig, time_step, plot_fig=False):
    # The FFT of the signal
    sig_fft = fftpack.fft(sig)
    # And the power (sig_fft is of complex dtype)
    power = np.abs(sig_fft)
    # The corresponding frequencies
    sample_freq = fftpack.fftfreq(sig.size, d=time_step)
    if plot_fig:
        # Plot the FFT power
        plt.figure(figsize=(6, 5))
        plt.plot(sample_freq, power)
        plt.xlabel('Frequency [Hz]')
        plt.ylabel('plower')
        # Find the peak frequency: we can focus on only the positive frequencies
        pos_mask = np.where(sample_freq > 0)
        freqs = sample_freq[pos_mask]
        peak_freq = freqs[power[pos_mask].argmax()]
        # An inner plot to show the peak frequency
        axes = plt.axes([0.55, 0.3, 0.3, 0.5])
        plt.title('Peak frequency')
        plt.plot(freqs[:8], power[:8])
        plt.setp(axes, yticks=[])
        # scipy.signal.find_peaks_cwt can also be used for more advanced
        # peak detection
    return sample_freq, power


def get_col_units(col_name, cols, cols_units):
    col_loc = cols.get_loc(col_name)
    col_units = cols_units[col_loc]
    return col_units


def multipage(filename, figs=None, dpi=200):
    pp = PdfPages(filename)
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for fig in figs:
        fig.savefig(pp, format='pdf')
    pp.close()


# def parallelize_dataframe(df, func, n_chunks, extra_args=()):
#     # df_split = np.array_split(df, 4)
#     # pool = Pool(4)
#     # df = pd.concat(pool.map(func, df_split))
#     # pool.close()
#     # pool.join()
#     with Pool(processes=n_chunks) as pool:
#         # starts the sub-processes without blocking
#         # pass the chunk to each worker process
#         proc_results = [pool.apply_async(func,
#                                          args=(chunk, extra_args))
#                         for chunk in proc_chunks]
#
#         # blocks until all results are fetched
#         result_chunks = [r.get() for r in proc_results]
#     return df

